/**************************************************************************
*                 X  J  D  I  C
*         Japanese-English Display program
*                                                   Author: Jim Breen
*
*  NOTE: This program is a port of the MS-DOS/Borland Graphics
*        version to run under X11. It needs to run from a "kterm" window
*        which is best set to its default JIS mode (i.e. not EUC or 
*        Shift-JIS). If you want kana keys, you will need to invoke
*        kinput.
***************************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/termio.h>

#define FALSE 0
#define TRUE 1
#define SEEK_END 2
#define MAXLINES 20	/* lines per display */
#define RADLINE 10	/* radicals/line in the "r" display */

struct stat *buf;

static enum
{
IOCTL_ORIG,
IOCTL_RAW,
} myi_status = IOCTL_ORIG;

static struct termio    orig,new;

int ShiftJIS = FALSE;
FILE *fpk,*fpki,*fpd,*fl,*fpi,*fpr;
unsigned char *dicbuff,*kanjbuff,*wskbuff;
unsigned long *jdxbuff,*kjdxbuff,*wjdxbuff;
unsigned char instr[256];
int Omode = 0,Smode = 0,Dmode = 0;
int maxlines = MAXLINES;
unsigned long kindlen,indlen,kdiclen,diclen,dbyte, dbyte2,hittab[1000];
int DispHit,hitind,FirstKanj = 0;
unsigned char kmodes[2][10] = {"ON","OFF"};
unsigned long chline,chpos,it,kindptr,indptr,lone,long0 = 0;
unsigned char tempout[80],sch[30],Bmsg[100];
unsigned char Rname[50] = {"radicals.tm"};
unsigned char IRname[50] = {"radicals.tm"};
unsigned char IDname[30] = {"edict"};
unsigned char IJDXname[30] = {"edict.xjdx"};
unsigned char IKDname[30] = {"kanjidic"};
unsigned char IKJDXname[30] = {"kanjidic.xjdx"};
unsigned char Dname[60] = {"edict"};
unsigned char JDXname[60] = {"edict.xjdx"};
unsigned char KDname[60] = {"kanjidic"};
unsigned char KJDXname[60] = {"kanjidic.xjdx"};
int jiver = 14;		/*The last time the index structure changed was Version1.4*/
unsigned char sver[] = {"1.0"};
unsigned char kanjidir[120] = {""};
unsigned char fbuff[512];
unsigned char Help[20][81] = {
"\n\nXJDIC USAGE SUMMARY\n\n",
"At the XJDIC SEARCH KEY: respond:\n\n",
"  a string of ascii, kana and/or kanji to look up in the main dictionary\n",
"  \\ to put it in Kanji Dictionary mode (see below)\n",
"  ? to get this display\n",
"  / to toggle the display between the two kanji_within_compound modes\n",
"  bye   to exit\n\n",
"In Kanji Dictionary mode, the prompt is KANJI LOOKUP TYPE:.  Respond\n\n",
"  k  followed by a single kanji or a kana reading (default)\n",
"  j  followed by the 4-digit hexadecimal JIS code for a kanji\n",
"  n  followed by the Nelson index number\n",
"  s  followed by a stroke count\n",
"  b  followed by either a radical (bushu) number or \"r\" to get a\n",
"     display of all radicals\n",
"  m  followed by an (English) kanji meaning\n",
"  r  initiates a display of all radicals and their numbers\n\n",
"See xjdic.man for full information\n",
"$$$"
};
unsigned char RVon[] = {0x1b,'[','7','m',0};
unsigned char RVoff[] = {0x1b,'[','m',0};


/*====== prototypes========================================================*/
void togglemode();
void DoKANJI();
void DoJIS();
void DoNumb(char *ch1, int rtyp);
void eight2seven(int *p1,int *p2);   /* Ken Lunde's routine  */
int addhit(long spot);
void GetEUC(unsigned char *eucline);
void cbreakon();
void cbreakoff();
void    ioctlorig();
void    ioctlraw();
void Lookup();
int Kstrcmpy(int slen);
void eight2seven(int *p1,int *p2);
unsigned long jindex(unsigned long it);
unsigned char dbchar(unsigned long it);
void DicInit();
void DicSet  ();
int Kstrcmp2();
/*====== end of prototypes========================================================*/
/* ==== ioctl routines supplied by Cameron Blackwood=======================*/
void    ioctlraw()
{
  if (myi_status == IOCTL_ORIG)
  {
    ioctl(0, TCGETA, &orig); ioctl(0, TCGETA, &new);
    new.c_lflag &= ~ICANON; new.c_lflag &= ~ISIG; new.c_lflag &= ~ECHO;
    new.c_lflag &= ~IXON;
    new.c_cc[4] = 1; new.c_cc[5] = 0;   ioctl(0, TCSETA, &new);
    myi_status=IOCTL_RAW;
  }
  else return;
}

void    ioctlorig()
{
  ioctl(0, TCSETA, &orig);
  myi_status = IOCTL_ORIG;
}

void cbreakoff()
{
	ioctlorig();
	/*system("stty -cbreak echo");*/
}

void cbreakon()
{
	ioctlraw();
	/*system("stty cbreak -echo");*/
}

/*======jis2sjis  (from Ken Lunde) =====================================*/
void jis2sjis(unsigned char *p1,unsigned char *p2) /* courtesy of Ken Lunde */
{
    register unsigned char c1 = *p1;
    register unsigned char c2 = *p2;
    register int rowOffset = c1 < 95 ? 112 : 176;
    register int cellOffset = c1 % 2 ? 31 + (c2 > 95) : 126;

    *p1 = ((c1 + 1) >> 1) + rowOffset;
    *p2 = c2 + cellOffset;
}
/*=====KOut===output kana/kanji according to selected mode===========*/
void KOut(unsigned char *sout)
{
	int i;
	unsigned char c1,c2;

	for (i = 0; i < strlen(sout); i++)
	{
		c1 = sout[i]; c2 = sout[i+1];
		if (c1 < 127)
		{
			printf("%c",c1);
			continue;
		}
		switch(Omode)
		{
		case 0 :  /* JIS (default)  */
			printf("%c$B%c%c%c(B",0x1b,c1&0x7f,c2&0x7f,0x1b);
			i++;
			break;

		case 1 : /* EUC (internal format) */
			printf("%c%c",c1,c2);
			i++;
			break;

		case 2 : /* Shift-JIS */
			c1 -= 128; c2 -= 128;
			jis2sjis(&c1,&c2);
			printf("%c%c",c1,c2);
			i++;
			break;
		}
	}
}
			
/*====DicSet check & load Dictionary and index files=======*/
void DicSet  ()
{
  	int i,isc,nodread;
  	lone = 1;

  	fpd=fopen(Dname,"rb");
  	if (fpd==NULL )
  	{
		printf("\nCannot open dictionary file: %s\n",Dname);
		exit(0);   
  	}
	buf = (void *)malloc(sizeof(struct stat));
	if(stat(Dname, buf) != 0)
	{
		/*perror();*/
		printf ("Stat() error\n");
		exit(0);
	}
	diclen = buf->st_size;

  	diclen++;
	dicbuff = (unsigned char *)calloc(diclen+100,1);
	if(dicbuff == NULL)
	{
		fprintf(stderr,"calloc() for dictionary failed.\n");
		fclose(fpd);
	exit(0);
	}
	nodread = diclen/1024;
	dbyte = fread((unsigned char *)dicbuff+1, 1024, nodread, fpd);
	nodread = diclen % 1024;
	dbyte = fread((unsigned char *)(dicbuff+(diclen/1024)*1024)+1, nodread,1, fpd);
	fclose(fpd);
	dicbuff[0] = 0xa;
  	dbyte = diclen;
  	fpi = fopen(JDXname,"rb");
  	if(fpi == NULL)
  	{
		printf("\nCannot open index file: %s\n",JDXname);
		exit(0);   
  	}
	if(stat(JDXname, buf) != 0)
        {
                /*perror();*/
		printf ("Stat() error\n");
                exit(0);
        }
        indlen = buf->st_size;
        jdxbuff = (unsigned long *)calloc(indlen+1024,1);
        if(jdxbuff == NULL)
        {
                fprintf(stderr,"calloc() for index failed.\n");
                fclose(fpi);
        exit(0);
        }


	indptr = indlen/4-1;
        nodread = indlen/1024+1;

	fread((long *)jdxbuff,1024,nodread,fpi);

	if (jdxbuff[0] != (diclen+jiver))
	{
		printf ("The dictionary and index files do not match! \n");
		exit(0);
	}

/* repeat for kanji dictionary */

  	fpk=fopen(KDname,"rb");
  	if (fpk==NULL )
  	{
		printf("\nCannot open Kanji dictionary file: %s\n",KDname);
		exit(0);   
  	}
	buf = (void *)malloc(sizeof(struct stat));
	if(stat(KDname, buf) != 0)
	{
		/*perror();*/
		printf ("Stat() error\n");
		exit(0);
	}
	kdiclen = buf->st_size;

  	kdiclen++;
	kanjbuff = (unsigned char *)calloc(kdiclen+100,1);
	if(kanjbuff == NULL)
	{
		fprintf(stderr,"calloc() for Kanji dictionary failed.\n");
		fclose(fpk);
		exit(0);
	}
	nodread = kdiclen/1024;
	dbyte = fread((unsigned char *)kanjbuff+1, 1024, nodread, fpk);
	nodread = kdiclen % 1024;
	dbyte = fread((unsigned char *)(kanjbuff+(kdiclen/1024)*1024)+1, nodread,1, fpk);
	fclose(fpk);
	kanjbuff[0] = 0xa;
  	dbyte = kdiclen;
  	fpki = fopen(KJDXname,"rb");
  	if(fpki == NULL)
  	{
		printf("\nCannot open kanji index file: %s\n",KJDXname);
		exit(0);   
  	}
	if(stat(KJDXname, buf) != 0)
        {
                /*perror();*/
		printf ("Stat() error\n");
                exit(0);
        }
        kindlen = buf->st_size;
        kjdxbuff = (unsigned long *)calloc(kindlen+1024,1);
        if(kjdxbuff == NULL)
        {
                fprintf(stderr,"calloc() for kanji index failed.\n");
                fclose(fpki);
        exit(0);
        }


	kindptr = kindlen/4-1;
        nodread = kindlen/1024+1;

	fread((long *)kjdxbuff,1024,nodread,fpki);

	if (kjdxbuff[0] != (kdiclen+jiver))
	{
		printf ("The kanji dictionary and index files do not match!\n");
		exit(0);
	}
}

/*=====dbchar====returns specified character from a dictionary===============*/

unsigned char dbchar(unsigned long it)
{
	if (Dmode == 0)
	{
		return(dicbuff[it]);
	}
	else
	{
		return(kanjbuff[it]);
	}
}

/*======jindex  returns specified entry from .xjdx file========*/


unsigned long jindex(unsigned long it)
{
	if (Dmode == 0)
	{
		return(jdxbuff[it]);
	}
	else
	{
		return(kjdxbuff[it]);

	}
}

/*=========string comparison function used in binary search (jbsearch)===*/
int Kstrcmp2(int slen)
{
	unsigned c1,c2;
	int i,rc1,rc2;
/* effectively does a strnicmp on two "strings" , one in fbuff               
   and the other pointed to by jindex(it)
   except it will make katakana and hiragana match (EUC A4 & A5) */

	for (i = 0; i<slen ; i++)
	{
		c1 = fbuff[i];
		c2 = dbchar(jindex(it)+i);
		if ((c1 == '\0')||(c2 == '\0')) return(0);
		if ((i % 2) ==0)
		{
			if (c1 == 0xA5)
			{
				c1 = 0xA4;
			}
			if (c2 == 0xA5)
			{
				c2 = 0xA4;
			}
		}
		if ((c1 >= 'A') && (c1 <= 'Z')) c1 |= 0x20; /*fix ucase*/
		if ((c2 >= 'A') && (c2 <= 'Z')) c2 |= 0x20;
		if (c1 != c2 ) 
		{
			rc1 = c1;
			rc2 = c2;
			return(rc1-rc2);
		}
	}
	return(0);
}
/*======function to test if this entry has already been displayed=====*/
int addhit(long spot)
{
	int i;
	for (i=0;i<=hitind;i++)
	{
		if(hittab[i] == spot) return(0);
	}
	hitind++;
	hittab[hitind] = spot;
	return(1);
}

/*====GetEUC - gets next text line, and converts to EUC ============*/

void GetEUC(unsigned char *eucline)
{
	int i,j,SI,p1,p2;

	eucline[0] = '\0';
	chline = 0;
	chpos = 0;
	SI = FALSE;
	j = 0;
	for (i = 0; i < strlen(instr); i++)
	{
		if((instr[i]==0x1b)&&(instr[i+1]=='$')&&(instr[i+2]=='B'))
		{
			SI = TRUE;
			i+=2;
			continue;
		}
		if((instr[i]==0x1b)&&(instr[i+1]=='$')&&(instr[i+2]=='@'))
		{
			SI = TRUE;
			i+=2;
			continue;
		}
		if((instr[i]==0x1b)&&(instr[i+1]=='(')&&(instr[i+2]=='J'))
		{
			SI = FALSE;
			i+=2;
			continue;
		}
		if((instr[i]==0x1b)&&(instr[i+1]=='(')&&(instr[i+2]=='B'))
		{
			SI = FALSE;
			i+=2;
			continue;
		}
		if (instr[i] == '\0')break;
		if (SI)
		{
			eucline[j] = instr[i] | 0x80;
		}
		else
		{
			eucline[j] = instr[i];
		}
		j++;
		eucline[j] = '\0';
	}
/*  fix up SHIFT-JIS, if present  */
	for (i = 0; i < strlen(eucline); i++)
	{
		p1 = eucline[i];
		if (p1 < 127)continue;
		p2 = eucline[i+1];
        	if ((p1 >= 129) && (p1 <= 159)) ShiftJIS = TRUE;
		if (((p1 >= 224) && (p1 <= 239))&& ((p2 >= 64) && (p2 <= 158))) ShiftJIS = TRUE;
		if (ShiftJIS)
		{
            		eight2seven(&p1,&p2);
            		p1 += 128;
            		p2 += 128;
			eucline[i] = p1;
			eucline[i+1] = p2;
		
		}
		i++;
	}
}

/*===eight2seven - convert ShiftJIS to JIS ===========================*/

void eight2seven(int *p1,int *p2)   /* Ken Lunde's routine  */
{
  int temp;

  temp = *p2;
  if ((*p2 >= 64) && (*p2 <= 158))
    *p2 -= 31;
  else if ((*p2 >= 159) && (*p2 <= 252))
    *p2 -= 126;
  if ((temp > 127) && (temp <= 158))
    (*p2)--;
  if ((*p1 >= 129) && (*p1 <= 159)) {
    if ((temp >= 64) && (temp <= 158))
      *p1 = ((*p1 - 113) * 2) + 1;
    else if ((temp >= 159) && (temp <= 252))
      *p1 = (*p1 - 112) * 2;
  }
  else if ((*p1 >= 224) && (*p1 <= 239)) {
    if ((temp >= 64) && (temp <= 158))
      *p1 = ((*p1 - 177) * 2) + 1;
    else if ((temp >= 159) && (temp <= 252))
      *p1 = (*p1 - 176) * 2;
  }
}

/*=========== Lookup - carry out dictionary search ================*/

void Lookup()
{
/*

Carries out a search for matches with the string "fbuff" in the specified
dictionary. It matches the full string, then progrssively shortens it.

*/
  	unsigned long endtest,schi,schix,schiy,schlen,lo,hi;
  	int dind,is,res,first,hit,skip,brace,klast=0;
	int ldisp,slench,slen,slenx,lrow=0,lcol=0,i;
  	unsigned int LUCont,khi,klo,cc;
  	unsigned long dt1,dt2,bshit[20];
  	unsigned char ops[20],c,resp,strtmp[30];

	/*  do a binary search through the index table looking for a match  */
	DispHit = FALSE;
	ldisp = 0;
  	khi = 0;
  	klo = 0;
  	lo = 1;
  	hi = indptr;
	endtest = indptr;
	if (Dmode != 0) 
	{
		hi = kindptr;
		endtest = kindptr;
	}
  	LUCont = TRUE;
	slench = 2;
	if (fbuff[0] < 127) slench = 1;
	slenx = strlen(fbuff)/slench;
	if (slenx > 20) slenx = 20;
	Smode = 0;
	if ((fbuff[0] == 0xa4)||(fbuff[0] == 0xa5)) Smode = 1;
  	for ( slen = 0; slen <slenx; slen++)
  	{
		while(1)
		{
  			it = (lo+hi)/2;
			res = Kstrcmp2(slen*slench+slench);
			if (res == 0)
			{
				bshit[slen] = it;
				break;
			}
			if (res < 0)
			{
				hi = it-1;
			}
			else
			{
				lo = it+1;
			}
			if (lo > hi)
			{
				LUCont = FALSE;
				break;
			}
		}
		if (!LUCont)
		{
			break;
		}
	}
  	if (slen == 0)
  	{
		printf("No Match Found\n");
		return;
  	}

  	hitind = 0;
  	hittab[0] = -1;
	/*  loop for each possible string length  */
	for (dind = slen-1; dind >= 0 ; dind--)
	{
	/*  scan backwards up the table to find the first   */
		it = bshit[dind];
		while (1)
		{
			if(Kstrcmp2(dind*slench+slench) == 0)
	/*		if(Kstrcmp2(it*slench+slench) == 0)*/
			{
				it--;
				if (it == 0)
				{
					it = 1;
					break;
				}
				continue;
			}
			else
			{
				it++;
				break;
			}
		}
  		hit = FALSE;

		/* this is the display loop - usually one line per entry  */
  		while (Kstrcmp2(dind*slench+slench) == 0)   /* display as long as there are matches*/
  		{
			schix = jindex(it);
			schiy = schix;
		/* back off to the start of this line   */
			while ((dbchar(schix) != 0xa) && (schix >= 0)) schix--;
			schix++;
		/* now work forwards, displaying the line  */
			if ((addhit(schix) != 0) && (!FirstKanj || (FirstKanj && (dbchar(schiy-1) < 127))))/* only display a line once  */
			{
			/* First, display the matched word/kana string  */
				DispHit = TRUE;
				if (Dmode == 0)
				{
					printf("%d: ",dind+1);
					while ( 1)
					{
						cc = dbchar(schiy);
						if (cc < 128)
						{
							if (!isalpha(cc)) break;
						}
						if (cc < 128) /* non-Japanese, display it*/
						{
							ops[0] = cc; ops[1] = 0;
							printf("%s",ops);
						}
						else   /* kana or kanji  */
						{
							if (khi == 0)
							{
								khi = cc;
							}
							else
							{
								klo = cc;
								ops[0] = khi; ops[1] = klo; ops[2] = 0;
								KOut(ops);
								khi = 0;
								klo = 0;
							}
						}
						schiy++;
	
					} /* end of match display  */
				}
				printf(" ");
				hit = FALSE;
				skip = FALSE;
				brace = FALSE;
				/* If the first word has already been displayed, skip it.
	   			Otherwise, put braces around the first word (kanji)  */
				if ((schix == jindex(it))&&(Dmode == 0))
				{
					skip = TRUE;
				}
				else
				{
					ops[0] = 0xa1; ops[1] = 0xda; ops[2] = 0; /* JIS [  */
					KOut(ops);
					brace = TRUE;
				}
				while ((cc = dbchar(schix)) != 0xa)
				{
					if (cc==' ')skip = FALSE;
					if ((brace)&&(cc==' '))
					{
						brace = FALSE;
						ops[0] = 0xa1; ops[1] = 0xdb; ops[2] = 0; /* JIS ]  */
						KOut(ops);
					}
					/* turn off yomikata skipping when a ']' is reached  */
					if ((skip)&&(cc==']'))
					{
						skip = FALSE;
						schix++;
						continue;
					}
					/* skip yomikata display during a kana search  */
					if ((Smode > 0)&&(cc=='[')) skip = TRUE;
					/* put a () around the yomikata in an ascii search */
					if ((Smode==0)&&(cc=='['))
					{
						ops[0] = 0xa1; ops[1] = 0xca; ops[2] = 0; /* JIS (  */
						KOut(ops);
						schix++;
						continue;
					}
					if ((Smode==0)&&(cc==']'))
					{
						ops[0] = 0xa1; ops[1] = 0xcb; ops[2] = 0; /* JIS )  */
						KOut(ops);
						schix++;
						continue;
					}
					if (skip)
					{
						schix++;
						continue;
					}
					if (cc < 128) /* non-Japanese, display it
								(fix up the MOKE "/") */
					{
						ops[0]= cc;
						ops[1] = '\0';
						if ((Dmode != 0) && (cc == '}')) 
	                                        {
                                                        ops[0] = ',';
                                        	}
						if ((Dmode != 0) && (cc == '{')) 
	                                        {
                                                	schix++;
                                                	continue;
                                        	}
						if (ops[0] == '/')
						{
							if (hit)
							{
								ops[0] = ',';
								ops[1] = ' ';
								ops[2] = '\0';
							}
							else
							{
								hit = TRUE;
								ops[0] = '\0';
							}
						}
						if ((dbchar(schix+1) == 0x0a) && (ops[0] == ',')) ops[0] = '\0';
						printf("%s",ops);
					}
					else   /* kana or kanji  */
					{
						if (khi == 0)
						{
							khi = cc;
						}
						else
						{
							klo = cc;
							ops[0] = khi ; ops[1] = klo ; ops[2] = 0;
							KOut(ops);
							khi = 0;
							klo = 0;
						}
					}
					schix++;
				}  /* end of line display loop  */
				printf("\n\r");
			}  /* of "once-per-line" test */
			else ldisp--;
			it++;
			if (it >= endtest) break;
			if(ldisp++ == maxlines-1)
			{
				printf("Continue displaying matches? (y/n)\n\r");
				ops[0] = getchar();
				fflush(stdin);
				if ((ops[0] == 'y')||(ops[0] == 'Y')) 
				{
					ldisp = 0;
					continue;
				}
				else 
				{
					if ((ops[0] != 'n') && (ops[0] != 'N') && (ops[0] != 0x0a) && (ops[0] != 0x0d)) ungetc(ops[0],stdin);
					return;
				}
			}
		}  /*  end of the "while it matches loop  */
		if (!DispHit) printf("No kanji compounds began with this character\n");
		if ((Dmode != 0) || (dind == 0)) return;
		printf("End of %d character matches. Continue for shorter matches? (y/n)\n\r",dind+1);
		ldisp = 0;
		ops[0] = getchar();
		fflush(stdin);
		if ((ops[0] == 'y')||(ops[0] == 'Y')) continue;
		if ((ops[0] != 'n')&&(ops[0] != 'N')) 
		{
			if ((ops[0] != 0x0a) && (ops[0] != 0x0d)) ungetc(ops[0],stdin);
		}
		break;
	}  /* end of the dind loop  */
}  /* end of lookup */

/*=====DoRADICALS===display Theresa's Radical File================*/

void DoRADICALS()
{

	int errf,i,j;
	unsigned char ops,rstr[20];

	fpr = fopen(Rname,"r");
	if (fpr == NULL)
	{
		printf("Sorry!, cannot open radicals.tm \n");
		return;
	}
	printf("\n RADICAL DISPLAY \n");
	fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
	j = 0;
	while(1)
	{
		errf = (fgets(rstr,19,fpr) == NULL);
		if (feof(fpr)||errf)
		{
			printf("\n");
			fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
			return;
		}
		while(rstr[strlen(rstr)-1] < 0x20) rstr[strlen(rstr)-1] = 0;
		if ((rstr[strlen(rstr)-2]  == ' ')&&(rstr[strlen(rstr)-1]  == '0'))
		{
			i = 0;
			j+=2;
			if (j >= maxlines-1)
			{
				printf("\nContinue displaying radicals? (y/n)\n");
				ops = getchar();
				fflush(stdin);
				if ((ops == 'y')||(ops == 'Y')) 
				{
					j = 0;
				}
				else 
				{
					if ((ops != 'n') && (ops != 'N') && (ops != 0x0a) && (ops != 0x0d)) ungetc(ops,stdin);
					return;
				}
			}
			if(rstr[0] > 127)
			{
				sprintf(tempout,"\n %c%c Stroke Radicals \n ",rstr[0],rstr[1]);
				KOut(tempout);
			}
			else
			{
				rstr[2] = 0;
				printf("\n %s Stroke Radicals \n ",rstr);
			}
			continue;
		}
		sprintf(tempout,"\t%s",rstr);
		KOut(tempout);
		i++;
		if (i == RADLINE)
		{
			i = 0;
			printf("\n ");
			j++;
			if (j >= maxlines-1)
			{
				printf("Continue displaying radicals? (y/n)\n\r");
				ops = getchar();
				fflush(stdin);
				if ((ops == 'y')||(ops == 'Y')) 
				{
					j = 0;
				}
				else 
				{
					if ((ops != 'n') && (ops != 'N') && (ops != 0x0a) && (ops != 0x0d)) ungetc(ops,stdin);
					return;
				}
			}
		}
	}
}

/*=======DuNumb === Common routine for kanji lookups==============================*/

void DoNumb(char *ch1, int rtyp)
{

	unsigned char ops,rstr[20];
	int i,j,k;

	cbreakoff();
	instr[0] = ch1[0];
	scanf("%s",instr+1);
	cbreakon();
	fflush(stdin);
	strcat(instr," ");
	Dmode =1;
	strcpy(fbuff,instr);
	fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
	Lookup();
	instr[0] = 0;
}

/*=========DoJIS === JIS kanji lookup==========================================*/

void DoJIS()
{
	int i;
	cbreakoff();
	scanf("%s",instr);
	cbreakon();
	fflush(stdin);
	for (i = 0;i <4; i++)
	{
		if ((instr[i] >= '0') && (instr[i] <= '9'))instr[i] = instr[i] - '0';
		if ((instr[i] >= 'a') && (instr[i] <= 'f'))instr[i] = instr[i]-'a'+0x0a;
	}
	instr[0] = ((instr[0] << 4) + instr[1]) | 0x80;
	instr[1] = ((instr[2] << 4) + instr[3]) | 0x80;
	instr[2] = 0;
	Dmode =1;
	strcpy(fbuff,instr);
	fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
	Lookup();
	instr[0] = 0;
}

/*=====  DoKANJI === Kanji single lookup ======================*/

void DoKANJI()
{
	int escf,i;
	unsigned char c;

	escf = FALSE;
	for (i = 0; (c != 0xd) && (c != 0xa); i++)
	{
		instr[i+1] = 0;
		c = getchar();
		if (c == 0x1b) escf = TRUE;
		instr[i] = c;
		if ((instr[i] == 'B')&&(instr[i-1] == '(')&&(instr[i-2] == 0x1b)) break;
	}
	fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
	printf(" %s\n\r",instr);
	fflush(stdin);
	GetEUC(fbuff);
	Dmode =1;
	Lookup();
	instr[0] = 0;
}
/*===togglemode===alternate between kanji compound modes===============*/
void togglemode()
{
	FirstKanj = (FirstKanj+1) % 2;
	printf("Display All Kanji Mode: %s\n",kmodes[FirstKanj]);
}
	

/*                  M A I N                                      */

main(argc,argv)
int argc;
char **argv;

{
	int i,ip,rubout,cmdmode,escf;
  	char *dicenv,strtmp[50];
  	unsigned char c;
	char xap[50];

  	printf("XJDIC Version %s (Japanese Dictionary) Copyright J.W.Breen 1992.\n",sver);

	if((dicenv = (char *)getenv("XJDIC")) != NULL)
	{
		strcpy (Rname,dicenv);
		strcat (Rname,"/");
		strcat (Rname,IRname);
		strcpy (Dname,dicenv);
		strcat (Dname,"/");
		strcat (Dname,IDname);
		strcpy (KDname,dicenv);
		strcat (KDname,"/");
		strcat (KDname,IKDname);
		strcpy (JDXname,dicenv);
		strcat (JDXname,"/");
		strcat (JDXname,IJDXname);
		strcpy (KJDXname,dicenv);
		strcat (KJDXname,"/");
		strcat (KJDXname,IKJDXname);
	}
/* process command-line options*/
	if (argc > 1)
	{
		for (i = 1; i < argc; i++)
		{
			strcpy(xap,argv[i]);
			if ((xap[0] == '-') && (xap[1] == 'j'))
			{
				if(xap[2] != 0)
				{
					strcpy(strtmp,xap+2);
				}
				else
				{
					i++;
					strcpy(xap,argv[i]);
					strcpy (strtmp,xap);
				}
				if (strtmp[0] == 's')
				{
					Omode = 2;
					printf("Output mode set to Shift-JIS\n");
				}
				if (strtmp[0] == 'e')
				{
					Omode = 1;
					printf("Output mode set to EUC\n");
				}
				continue;
			}
			if ((xap[0] == '-') && (xap[1] == 'l'))
			{
				if(xap[2] != 0)
				{
					strcpy(strtmp,xap+2);
				}
				else
				{
					i++;
					strcpy(xap,argv[i]);
					strcpy (strtmp,xap);
				}
				maxlines = atoi(strtmp);
				if(maxlines < 5) maxlines = 5;
				printf("Setting lines per display to %d\n",maxlines);
				continue;
			}
			if ((xap[0] == '-') && (xap[1] == 'd'))
			{
				if(xap[2] != 0)
				{
					strcpy(strtmp,xap+2);
				}
				else
				{
					i++;
					strcpy(xap,argv[i]);
					strcpy (strtmp,xap);
				}
				strcpy (Dname,strtmp);
				strcpy (JDXname,strtmp);
				strcat (JDXname,".xjdx");
				printf("Command-line request to use dictionary files: %s and %s\n",Dname,JDXname);
				continue;
			}
			if ((xap[0] == '-') && (xap[1] == 'k'))
			{
				if(xap[2] != 0)
				{
					strcpy(strtmp,xap+2);
				}
				else
				{
					i++;
					strcpy(xap,argv[i]);
					strcpy (strtmp,xap);
				}
				strcpy (KDname,strtmp);
				strcpy (KJDXname,strtmp);
				strcat (KJDXname,".xjdx");
				printf("Command-line request to use kanji dictionary files: %s and %s\n",Dname,JDXname);
				continue;
			}
			if ((xap[0] == '-') && (xap[1] == 'h'))
			{
printf("\nUSAGE: xjdic <options>\n\nwhere options are:\n\n");
printf("   -h this information\n   -d dictionary file_path\n   -k kanji-dictionary file_path\n   -l nn (where nn is lines per display)\n");
printf("   -j x (where x is e for EUC Japanese codes, and s for Shift-JIS)\n");
printf("\nSee xjdic.man for full information\n");
exit(0);
			}
		}
	}
  	printf ("Loading Dictionary and Index files.  Please wait.....\n");
  	DicSet ();      /* set up dictionary buffers  */
	togglemode();
	cbreakon();
	printf("\n(Enter ? for a summary of operating instructions)\n");

  	while (1)
	{
		printf("\n\r%sXJDIC SEARCH KEY:%s ",RVon,RVoff);
		c = 0;
		rubout = FALSE;
		cmdmode = FALSE;
		escf = FALSE;
		for (i = 0; (c != 0xd) && (c != 0xa); i++)
		{
			instr[i+1] = 0;
			c = getchar();
			if (c == 0x1b) escf = TRUE;
			if ((c == '?') && (!escf))
			{
				for (ip = 0; strcmp(Help[ip],"$$$")!=0;ip++) printf("%s",Help[ip]);
				break;
			}
			if ((c == '/') && (!escf))
			{
				cmdmode = TRUE; 
				break;
			}
			if ((c == '\\') && (!escf))
			{
				cmdmode = TRUE; 
				break;
			}
			if (c == 0x7f) 
			{
				rubout = TRUE; 
				break;
			}
			instr[i] = c;
			if (!escf)printf("%c",c);
			if ((instr[i] == 'B')&&(instr[i-1] == '(')&&(instr[i-2] == 0x1b)) break;
		}
		fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
		if ((instr[2] == 'e')&&(instr[1] == 'y')&&(instr[0] == 'b')) 
		{
			cbreakoff();
			exit(0);
		}
		if(rubout) continue;
		if(cmdmode)
		{
			if (c == '/')
			{
				printf("\r                                      \r");
				togglemode();
				continue;
			}
			printf("\r                                      \r");
			printf("%sKANJI LOOKUP TYPE:%s ",RVon,RVoff);
			c = getchar();

			switch (c)
			{
				case 'n' :
				case 'N' :
					
					printf("\r                                      \r");
					printf("%sNELSON NUMBER:%s",RVon,RVoff);
					DoNumb("N",0);
					break;
				case 'b' :
				case 'B' :
					printf("\r                                      \r");
					printf("%sBUSHU NUMBER:%s",RVon,RVoff);
					DoNumb("B",1);
					break;
				case 'm' :
				case 'M' :
					printf("\r                                      \r");
					printf("%sMEANING:%s",RVon,RVoff);
					cbreakoff();
					scanf("%s",instr);
					cbreakon();
					fflush(stdin);
					Dmode =1;
					strcpy(fbuff,instr);
					fseek(stdin,0L,SEEK_END); /*kill any leftovers*/
					Lookup();
					instr[0] = 0;
					break;
				case 's' :
				case 'S' :
					printf("\r                                      \r");
					printf("%sSTROKE COUNT:%s",RVon,RVoff);
					DoNumb("S",1);
					break;
				case 'j' :
				case 'J' :
					printf("\r                                      \r");
					printf("%sJIS CODE:%s",RVon,RVoff);
					DoJIS();
					break;
				case 'r' :
				case 'R' :
					DoRADICALS();
					break;
				case 'k' :
				case 'K' :
					printf("\r                                      \r");
					printf("%sKANJI/KANA:%s",RVon,RVoff);
					DoKANJI();
					break;
				default :
					ungetc(c,stdin);
					printf("\r                                      \r");
					printf("%sASSUMING KANJI/KANA:%s",RVon,RVoff);
					DoKANJI();
					break;
			}
		}
		if(strlen(instr) < 2) continue;
		GetEUC(fbuff);
		sprintf(tempout,"\nSearching for: %s%s%s\n",RVon,fbuff,RVoff);
		KOut(tempout);
		Dmode = 0;
		Lookup();
	}
}
